@@ -5,25 +5,30 @@ use crate::{
     errors::{FeatureBuildStage, FeatureGraphWarning},
     graph::{
         feature::{
-            ConditionalLinkImpl, FeatureEdge, FeatureGraphImpl, FeatureMetadataImpl, FeatureNode,
-            FeatureType,
+            ConditionalLinkImpl, FeatureEdge, FeatureGraphImpl, FeatureLabel, FeatureMetadataImpl,
+            FeatureNode, WeakDependencies, WeakIndex,
         },
-        DepRequiredOrOptional, DependencyReq, FeatureIx, PackageGraph, PackageIx, PackageLink,
-        PackageMetadata,
+        DepRequiredOrOptional, DependencyReq, FeatureIndexInPackage, FeatureIx, NamedFeatureDep,
+        PackageGraph, PackageIx, PackageLink, PackageMetadata,
     },
     platform::PlatformStatusImpl,
 };
 use cargo_metadata::DependencyKind;
 use once_cell::sync::OnceCell;
-use petgraph::prelude::*;
+use petgraph::{prelude::*, visit::IntoEdgeReferences};
+use smallvec::SmallVec;
 use std::{collections::HashMap, iter};
 
+pub(super) type FeaturePetgraph = Graph<FeatureNode, FeatureEdge, Directed, FeatureIx>;
+pub(super) type FeatureEdgeReference<'g> = <&'g FeaturePetgraph as IntoEdgeReferences>::EdgeRef;
+
 #[derive(Debug)]
 pub(super) struct FeatureGraphBuildState {
-    graph: Graph<FeatureNode, FeatureEdge, Directed, FeatureIx>,
+    graph: FeaturePetgraph,
     // Map from package ixs to the base (first) feature for each package.
     base_ixs: Vec<NodeIndex<FeatureIx>>,
     map: HashMap<FeatureNode, FeatureMetadataImpl>,
+    weak: WeakDependencies,
     warnings: Vec<FeatureGraphWarning>,
 }
 
@@ -37,6 +42,7 @@ impl FeatureGraphBuildState {
             // the end.
             base_ixs: Vec::with_capacity(package_count + 1),
             map: HashMap::with_capacity(package_count),
+            weak: WeakDependencies::new(),
             warnings: vec![],
         }
     }
@@ -45,22 +51,15 @@ impl FeatureGraphBuildState {
     /// feature to the base package.
     pub(super) fn add_nodes(&mut self, package: PackageMetadata<'_>) {
         let base_node = FeatureNode::base(package.package_ix());
-        let base_ix = self.add_node(base_node, FeatureType::BasePackage);
+        let base_ix = self.add_node(base_node);
         self.base_ixs.push(base_ix);
-        FeatureNode::named_features(package).for_each(|feature_node| {
-            let feature_ix = self.add_node(feature_node, FeatureType::NamedFeature);
-            self.graph
-                .update_edge(feature_ix, base_ix, FeatureEdge::FeatureToBase);
-        });
-
-        package.optional_deps_full().for_each(|(n, _)| {
-            let dep_idx = self.add_node(
-                FeatureNode::new(package.package_ix(), n),
-                FeatureType::OptionalDep,
-            );
-            self.graph
-                .update_edge(dep_idx, base_ix, FeatureEdge::FeatureToBase);
-        });
+        FeatureNode::named_features(package)
+            .chain(FeatureNode::optional_deps(package))
+            .for_each(|feature_node| {
+                let feature_ix = self.add_node(feature_node);
+                self.graph
+                    .update_edge(feature_ix, base_ix, FeatureEdge::FeatureToBase);
+            });
     }
 
     /// Mark the end of adding nodes.
@@ -81,64 +80,12 @@ impl FeatureGraphBuildState {
                 let to_nodes_edges: Vec<_> = feature_deps
                     .iter()
                     .flat_map(|feature_dep| {
-                        let (dep_name, to_feature) = Self::split_feature_dep(feature_dep);
-                        let (cross_node_edge, same_node_edge) = match dep_name {
-                            Some(dep_name) => {
-                                if let Some(link) = dep_name_to_link.get(dep_name) {
-                                    // dependency from (`main`, `a`) to (`dep, `foo`)
-                                    let cross_node_edge = self
-                                        .make_named_feature_node(
-                                            &metadata,
-                                            from_feature,
-                                            &link.to(),
-                                            to_feature,
-                                            true,
-                                        )
-                                        .map(|cross_node| {
-                                            // This is a cross-package link. The platform-specific
-                                            // requirements still apply, so grab them from the
-                                            // PackageLink.
-                                            (cross_node, Self::make_named_feature_cross_edge(link))
-                                        });
-
-                                    // If the package is present as an optional dependency, it is
-                                    // implicitly activated by the feature:
-                                    // from (`main`, `a`) to (`main`, `dep`)
-                                    let same_node_edge = self
-                                        .make_named_feature_node(
-                                            &metadata,
-                                            from_feature,
-                                            &metadata,
-                                            dep_name,
-                                            // Don't warn if this dep isn't optional.
-                                            false,
-                                        )
-                                        .map(|same_node| {
-                                            (same_node, Self::make_named_feature_cross_edge(link))
-                                        });
-                                    (cross_node_edge, same_node_edge)
-                                } else {
-                                    // The destination package was unknown to the graph.
-                                    // XXX may need to be revisited if we start modeling unresolved
-                                    // dependencies.
-                                    (None, None)
-                                }
-                            }
-                            None => {
-                                let same_node_edge = self
-                                    .make_named_feature_node(
-                                        &metadata,
-                                        from_feature,
-                                        &metadata,
-                                        to_feature,
-                                        true,
-                                    )
-                                    .map(|same_node| (same_node, FeatureEdge::FeatureDependency));
-                                (None, same_node_edge)
-                            }
-                        };
-
-                        cross_node_edge.into_iter().chain(same_node_edge)
+                        self.nodes_for_named_feature_dep(
+                            metadata,
+                            FeatureLabel::Named(from_feature),
+                            feature_dep,
+                            &dep_name_to_link,
+                        )
                     })
                     // The flat_map above holds an &mut reference to self, which is why it needs to
                     // be collected.
@@ -150,29 +97,121 @@ impl FeatureGraphBuildState {
             })
     }
 
-    /// Split a feature dep into package and feature names.
-    ///
-    /// "foo" -> (None, "foo")
-    /// "dep/foo" -> (Some("dep"), "foo")
-    fn split_feature_dep(feature_dep: &str) -> (Option<&str>, &str) {
-        let mut rsplit = feature_dep.rsplitn(2, '/');
-        let to_feature_name = rsplit
-            .next()
-            .expect("rsplitn should return at least one element");
-        let dep_name = rsplit.next();
+    fn nodes_for_named_feature_dep(
+        &mut self,
+        metadata: PackageMetadata<'_>,
+        from_label: FeatureLabel<'_>,
+        feature_dep: &NamedFeatureDep,
+        dep_name_to_link: &HashMap<&str, PackageLink>,
+    ) -> SmallVec<[(FeatureNode, FeatureEdge); 3]> {
+        let mut nodes_edges: SmallVec<[(FeatureNode, FeatureEdge); 3]> = SmallVec::new();
 
-        (dep_name, to_feature_name)
+        match feature_dep {
+            NamedFeatureDep::DependencyNamedFeature {
+                dep_name,
+                feature,
+                weak,
+            } => {
+                if let Some(link) = dep_name_to_link.get(dep_name.as_ref()) {
+                    let weak_index = weak.then(|| self.weak.insert(link.edge_ix()));
+
+                    // Dependency from (`main`, `a`) to (`dep, `foo`)
+                    if let Some(cross_node) = self.make_named_feature_node(
+                        &metadata,
+                        from_label,
+                        &link.to(),
+                        FeatureLabel::Named(feature.as_ref()),
+                        true,
+                    ) {
+                        // This is a cross-package link. The platform-specific
+                        // requirements still apply, so grab them from the
+                        // PackageLink.
+                        nodes_edges.push((
+                            cross_node,
+                            Self::make_named_feature_cross_edge(link, weak_index),
+                        ));
+                    };
+
+                    // If the package is present as an optional dependency, it is
+                    // implicitly activated by the feature:
+                    // from (`main`, `a`) to (`main`, `dep:dep`)
+                    if let Some(same_node) = self.make_named_feature_node(
+                        &metadata,
+                        from_label,
+                        &metadata,
+                        FeatureLabel::OptionalDependency(dep_name),
+                        // Don't warn if this dep isn't optional.
+                        false,
+                    ) {
+                        nodes_edges.push((
+                            same_node,
+                            Self::make_named_feature_cross_edge(link, weak_index),
+                        ));
+                    }
+
+                    // Finally, (`main`, `a`) to (`main`, `dep`) -- if this is a non-weak dependency
+                    // and a named feature by this name is present, it also gets activated (even if
+                    // the named feature has no relation to the optional dependency).
+                    if !*weak {
+                        if let Some(same_named_feature_node) = self.make_named_feature_node(
+                            &metadata,
+                            from_label,
+                            &metadata,
+                            FeatureLabel::Named(dep_name),
+                            // Don't warn if this dep isn't optional.
+                            false,
+                        ) {
+                            nodes_edges.push((
+                                same_named_feature_node,
+                                Self::make_named_feature_cross_edge(link, None),
+                            ));
+                        }
+                    }
+                }
+            }
+            NamedFeatureDep::NamedFeature(feature_name) => {
+                if let Some(same_node) = self.make_named_feature_node(
+                    &metadata,
+                    from_label,
+                    &metadata,
+                    FeatureLabel::Named(feature_name.as_ref()),
+                    true,
+                ) {
+                    nodes_edges.push((same_node, FeatureEdge::NamedFeature));
+                }
+            }
+            NamedFeatureDep::OptionalDependency(dep_name) => {
+                if let Some(same_node) = self.make_named_feature_node(
+                    &metadata,
+                    from_label,
+                    &metadata,
+                    FeatureLabel::OptionalDependency(dep_name.as_ref()),
+                    true,
+                ) {
+                    if let Some(link) = dep_name_to_link.get(dep_name.as_ref()) {
+                        nodes_edges.push((
+                            same_node,
+                            FeatureEdge::NamedFeatureDepColon(
+                                Self::make_full_conditional_link_impl(link),
+                            ),
+                        ));
+                    }
+                }
+            }
+        };
+
+        nodes_edges
     }
 
     fn make_named_feature_node(
         &mut self,
         from_package: &PackageMetadata<'_>,
-        from_feature: &str,
+        from_label: FeatureLabel<'_>,
         to_package: &PackageMetadata<'_>,
-        to_feature: &str,
+        to_label: FeatureLabel<'_>,
         warn: bool,
     ) -> Option<FeatureNode> {
-        match to_package.get_feature_idx(to_feature) {
+        match to_package.get_feature_idx(to_label) {
             Some(idx) => Some(FeatureNode::new(to_package.package_ix(), idx)),
             None => {
                 // It is possible to specify a feature that doesn't actually exist, and cargo will
@@ -187,10 +226,10 @@ impl FeatureGraphBuildState {
                     self.warnings.push(FeatureGraphWarning::MissingFeature {
                         stage: FeatureBuildStage::AddNamedFeatureEdges {
                             package_id: from_package.id().clone(),
-                            from_feature: from_feature.to_string(),
+                            from_feature: from_label.to_string(),
                         },
                         package_id: to_package.id().clone(),
-                        feature_name: to_feature.to_string(),
+                        feature_name: to_label.to_string(),
                     });
                 }
                 None
@@ -207,9 +246,23 @@ impl FeatureGraphBuildState {
     ///
     /// (a link (`from`, `a`) to (`dep`, `foo`) is created.
     ///
-    /// If `dep` is optional, the edge (`from`, `a`) to (`from`, `dep`) is also a `CrossPackage`
-    /// edge.
-    fn make_named_feature_cross_edge(link: &PackageLink<'_>) -> FeatureEdge {
+    /// If `dep` is optional, the edge (`from`, `a`) to (`from`, `dep`) is also a
+    /// `NamedFeatureWithSlash` edge.
+    fn make_named_feature_cross_edge(
+        link: &PackageLink<'_>,
+        weak_index: Option<WeakIndex>,
+    ) -> FeatureEdge {
+        // This edge is enabled if the feature is enabled, which means the union of (required,
+        // optional) build conditions.
+        FeatureEdge::NamedFeatureWithSlash {
+            link: Self::make_full_conditional_link_impl(link),
+            weak_index,
+        }
+    }
+
+    // Creates a "full" conditional link, unifying requirements across all dependency lines.
+    // This should not be used in add_dependency_edges below!
+    fn make_full_conditional_link_impl(link: &PackageLink<'_>) -> ConditionalLinkImpl {
         // This edge is enabled if the feature is enabled, which means the union of (required,
         // optional) build conditions.
         fn combine_req_opt(req: DependencyReq<'_>) -> PlatformStatusImpl {
@@ -218,12 +271,12 @@ impl FeatureGraphBuildState {
             required
         }
 
-        FeatureEdge::Conditional(ConditionalLinkImpl {
+        ConditionalLinkImpl {
             package_edge_ix: link.edge_ix(),
             normal: combine_req_opt(link.normal()),
             build: combine_req_opt(link.build()),
             dev: combine_req_opt(link.dev()),
-        })
+        }
     }
 
     pub(super) fn add_dependency_edges(&mut self, link: PackageLink<'_>) {
@@ -289,31 +342,23 @@ impl FeatureGraphBuildState {
             // package metadata.
             let from_node = FeatureNode::new(
                 from.package_ix(),
-                from.get_feature_idx(link.dep_name()).unwrap_or_else(|| {
-                    panic!(
+                from.get_feature_idx(FeatureLabel::OptionalDependency(link.dep_name()))
+                    .unwrap_or_else(|| {
+                        panic!(
                         "while adding feature edges, for package '{}', optional dep '{}' missing",
                         from.id(),
                         link.dep_name(),
                     );
-                }),
+                    }),
             );
             self.add_edges(from_node, optional_req.finish());
         }
     }
 
-    fn add_node(
-        &mut self,
-        feature_id: FeatureNode,
-        feature_type: FeatureType,
-    ) -> NodeIndex<FeatureIx> {
+    fn add_node(&mut self, feature_id: FeatureNode) -> NodeIndex<FeatureIx> {
         let feature_ix = self.graph.add_node(feature_id);
-        self.map.insert(
-            feature_id,
-            FeatureMetadataImpl {
-                feature_ix,
-                feature_type,
-            },
-        );
+        self.map
+            .insert(feature_id, FeatureMetadataImpl { feature_ix });
         feature_ix
     }
 
@@ -351,6 +396,11 @@ impl FeatureGraphBuildState {
                     // "dep/feat" causes a cross link to be created from "main/feat" to "main/dep".
                     // However, the "dep" encountered later upgrades this link to a feature
                     // dependency.
+                    //
+                    // This could also be an upgrade from a weak to a non-weak dependency:
+                    //
+                    // [features]
+                    // feat = ["dep?/feat", "dep/feat2"]
                     let old_edge = self
                         .graph
                         .edge_weight_mut(edge_ix)
@@ -358,12 +408,40 @@ impl FeatureGraphBuildState {
                     #[allow(clippy::single_match)]
                     match (old_edge, edge) {
                         (
-                            old_edge @ FeatureEdge::Conditional(_),
-                            edge @ FeatureEdge::FeatureDependency,
+                            FeatureEdge::NamedFeatureWithSlash {
+                                weak_index: old_weak_index,
+                                ..
+                            },
+                            FeatureEdge::NamedFeatureWithSlash { weak_index, .. },
                         ) => {
-                            // Upgrade this edge.
+                            if old_weak_index.is_some() && weak_index.is_some() {
+                                debug_assert_eq!(
+                                    *old_weak_index, weak_index,
+                                    "weak indexes should match if some"
+                                );
+                            }
+                            // Upgrade this edge from weak to non-weak.
+                            if weak_index.is_none() {
+                                *old_weak_index = None;
+                            }
+                        }
+                        (
+                            old_edge @ FeatureEdge::NamedFeatureWithSlash { .. },
+                            edge @ FeatureEdge::NamedFeature
+                            | edge @ FeatureEdge::NamedFeatureDepColon(_),
+                        ) => {
+                            // Upgrade this edge from / conditional to dep: conditional or unconditional.
                             *old_edge = edge;
                         }
+                        (
+                            old_edge @ FeatureEdge::NamedFeatureDepColon(_),
+                            edge @ FeatureEdge::NamedFeature,
+                        ) => {
+                            // Upgrade this edge from dep: conditional to unconditional.
+                            // XXX: can this ever happen?
+                            *old_edge = edge;
+                        }
+
                         _ => {
                             // In all other cases, leave the old edge alone.
                         }
@@ -387,6 +465,7 @@ impl FeatureGraphBuildState {
             map: self.map,
             warnings: self.warnings,
             sccs: OnceCell::new(),
+            weak: self.weak,
         }
     }
 }
@@ -396,9 +475,9 @@ struct FeatureReq<'g> {
     link: PackageLink<'g>,
     to: PackageMetadata<'g>,
     edge_ix: EdgeIndex<PackageIx>,
-    to_default_idx: Option<usize>,
+    to_default_idx: FeatureIndexInPackage,
     // This will contain any build states that aren't empty.
-    features: HashMap<Option<usize>, DependencyBuildState>,
+    features: HashMap<FeatureIndexInPackage, DependencyBuildState>,
 }
 
 impl<'g> FeatureReq<'g> {
@@ -408,7 +487,9 @@ impl<'g> FeatureReq<'g> {
             link,
             to,
             edge_ix: link.edge_ix(),
-            to_default_idx: to.get_feature_idx("default"),
+            to_default_idx: to
+                .get_feature_idx(FeatureLabel::Named("default"))
+                .unwrap_or(FeatureIndexInPackage::Base),
             features: HashMap::new(),
         }
     }
@@ -425,14 +506,14 @@ impl<'g> FeatureReq<'g> {
         warnings: &mut Vec<FeatureGraphWarning>,
     ) {
         // Base feature.
-        self.extend(None, dep_kind, &req.build_if);
+        self.extend(FeatureIndexInPackage::Base, dep_kind, &req.build_if);
         // Default feature (or base if it isn't present).
         self.extend(self.to_default_idx, dep_kind, &req.default_features_if);
 
         for (feature, status) in &req.feature_targets {
-            match self.to.get_feature_idx(feature) {
+            match self.to.get_feature_idx(FeatureLabel::Named(feature)) {
                 Some(feature_idx) => {
-                    self.extend(Some(feature_idx), dep_kind, status);
+                    self.extend(feature_idx, dep_kind, status);
                 }
                 None => {
                     // The destination feature is missing -- this is accepted by cargo
@@ -452,7 +533,7 @@ impl<'g> FeatureReq<'g> {
 
     fn extend(
         &mut self,
-        feature_idx: Option<usize>,
+        feature_idx: FeatureIndexInPackage,
         dep_kind: DependencyKind,
         status: &PlatformStatusImpl,
     ) {
@@ -473,7 +554,7 @@ impl<'g> FeatureReq<'g> {
                 // extend ensures that the build states aren't empty. Double-check that.
                 debug_assert!(!build_state.is_empty(), "build states are always non-empty");
                 (
-                    FeatureNode::new_opt(package_ix, feature_idx),
+                    FeatureNode::new(package_ix, feature_idx),
                     build_state.finish(),
                 )
             })
@@ -512,7 +593,7 @@ impl DependencyBuildState {
     }
 
     fn finish(self) -> FeatureEdge {
-        FeatureEdge::Conditional(ConditionalLinkImpl {
+        FeatureEdge::DependenciesSection(ConditionalLinkImpl {
             package_edge_ix: self.package_edge_ix,
             normal: self.normal,
             build: self.build,
