@@ -5,8 +5,12 @@ use crate::{
     debug_ignore::DebugIgnore,
     errors::FeatureGraphWarning,
     graph::{
-        feature::{build::FeatureGraphBuildState, Cycles, FeatureFilter, FeatureList},
-        DependencyDirection, FeatureIx, PackageGraph, PackageIx, PackageLink, PackageMetadata,
+        feature::{
+            build::{FeatureGraphBuildState, FeaturePetgraph},
+            Cycles, FeatureFilter, FeatureList, WeakDependencies, WeakIndex,
+        },
+        DependencyDirection, FeatureIndexInPackage, FeatureIx, PackageGraph, PackageIx,
+        PackageLink, PackageMetadata,
     },
     petgraph_support::{scc::Sccs, topo::TopoWithCycles},
     platform::{PlatformStatus, PlatformStatusImpl},
@@ -96,6 +100,12 @@ impl<'g> FeatureGraph<'g> {
         self.dep_graph().edge_count()
     }
 
+    /// Returns true if this feature graph contains the specified feature.
+    pub fn contains(&self, feature_id: impl Into<FeatureId<'g>>) -> bool {
+        let feature_id = feature_id.into();
+        FeatureNode::from_id(self, feature_id).is_some()
+    }
+
     /// Returns metadata for the given feature ID, or `None` if the feature wasn't found.
     pub fn metadata(
         &self,
@@ -219,8 +229,10 @@ impl<'g> FeatureGraph<'g> {
         self.inner.sccs.get_or_init(|| {
             let edge_filtered =
                 EdgeFiltered::from_fn(self.dep_graph(), |edge| match edge.weight() {
-                    FeatureEdge::Conditional(conditional_link) => !conditional_link.dev_only(),
-                    FeatureEdge::FeatureDependency | FeatureEdge::FeatureToBase => true,
+                    FeatureEdge::DependenciesSection(link)
+                    | FeatureEdge::NamedFeatureDepColon(link)
+                    | FeatureEdge::NamedFeatureWithSlash { link, .. } => !link.dev_only(),
+                    FeatureEdge::NamedFeature | FeatureEdge::FeatureToBase => true,
                 });
             // Sort the entire graph without dev-only edges -- a correct graph would be cycle-free
             // but we don't currently do a consistency check for this so handle cycles.
@@ -259,7 +271,7 @@ impl<'g> FeatureGraph<'g> {
         self.inner.map.get(node)
     }
 
-    pub(super) fn dep_graph(&self) -> &'g Graph<FeatureNode, FeatureEdge, Directed, FeatureIx> {
+    pub(super) fn dep_graph(&self) -> &'g FeaturePetgraph {
         &self.inner.graph
     }
 
@@ -270,12 +282,21 @@ impl<'g> FeatureGraph<'g> {
         target_ix: NodeIndex<FeatureIx>,
         edge_ix: EdgeIndex<FeatureIx>,
         edge: Option<&'g FeatureEdge>,
-    ) -> Option<ConditionalLink<'g>> {
-        match edge.unwrap_or_else(|| &self.dep_graph()[edge_ix]) {
-            FeatureEdge::FeatureDependency | FeatureEdge::FeatureToBase => None,
-            FeatureEdge::Conditional(inner) => Some(ConditionalLink::new(
-                *self, source_ix, target_ix, edge_ix, inner,
-            )),
+    ) -> Option<(ConditionalLink<'g>, Option<WeakIndex>)> {
+        let edge = edge.unwrap_or_else(|| &self.dep_graph()[edge_ix]);
+
+        match edge {
+            FeatureEdge::NamedFeature | FeatureEdge::FeatureToBase => None,
+            FeatureEdge::DependenciesSection(link) | FeatureEdge::NamedFeatureDepColon(link) => {
+                let link = ConditionalLink::new(*self, source_ix, target_ix, edge_ix, link);
+                // Dependency section and dep:foo style conditional links are always non-weak.
+                let weak_index = None;
+                Some((link, weak_index))
+            }
+            FeatureEdge::NamedFeatureWithSlash { link, weak_index } => {
+                let link = ConditionalLink::new(*self, source_ix, target_ix, edge_ix, link);
+                Some((link, *weak_index))
+            }
         }
     }
 
@@ -294,7 +315,10 @@ impl<'g> FeatureGraph<'g> {
     ) -> bool {
         // Filter out conditional edges.
         let edge_filtered = EdgeFiltered::from_fn(self.dep_graph(), |edge_ref| {
-            !matches!(edge_ref.weight(), FeatureEdge::Conditional(_))
+            !matches!(
+                edge_ref.weight(),
+                FeatureEdge::DependenciesSection(_) | FeatureEdge::NamedFeatureWithSlash { .. }
+            )
         });
         has_path_connecting(&edge_filtered, a_ix, b_ix, None)
     }
@@ -383,25 +407,39 @@ impl<'g> FeatureGraph<'g> {
 #[derive(Clone, Copy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct FeatureId<'g> {
     package_id: &'g PackageId,
-    feature: Option<&'g str>,
+    label: FeatureLabel<'g>,
 }
 
 assert_covariant!(FeatureId);
 
 impl<'g> FeatureId<'g> {
-    /// Creates a new `FeatureId`.
-    pub fn new(package_id: &'g PackageId, feature: &'g str) -> Self {
+    /// Creates a new `FeatureId` with the given [`PackageId`] and [`FeatureLabel`].
+    pub fn new(package_id: &'g PackageId, label: FeatureLabel<'g>) -> Self {
+        Self { package_id, label }
+    }
+
+    /// Creates a new `FeatureId` representing a named feature in the `[features]` section,
+    /// or an implicit named feature created by an optional dependency.
+    pub fn named(package_id: &'g PackageId, feature_name: &'g str) -> Self {
         Self {
             package_id,
-            feature: Some(feature),
+            label: FeatureLabel::Named(feature_name),
         }
     }
 
-    /// Creates a new `FeatureId` representing the "base" feature for a package.
+    /// Creates a new `FeatureId` representing an optional dependency.
+    pub fn optional_dependency(package_id: &'g PackageId, dep_name: &'g str) -> Self {
+        Self {
+            package_id,
+            label: FeatureLabel::OptionalDependency(dep_name),
+        }
+    }
+
+    /// Creates a new `FeatureId` representing the base feature for a package.
     pub fn base(package_id: &'g PackageId) -> Self {
         Self {
             package_id,
-            feature: None,
+            label: FeatureLabel::Base,
         }
     }
 
@@ -413,15 +451,15 @@ impl<'g> FeatureId<'g> {
         let feature = Self::node_to_feature(metadata, node);
         Self {
             package_id,
-            feature,
+            label: feature,
         }
     }
 
     pub(super) fn node_to_feature(
         metadata: PackageMetadata<'g>,
         node: &FeatureNode,
-    ) -> Option<&'g str> {
-        Some(metadata.feature_idx_to_name(node.feature_idx?))
+    ) -> FeatureLabel<'g> {
+        metadata.feature_idx_to_label(node.feature_idx)
     }
 
     /// Returns the package ID.
@@ -429,42 +467,41 @@ impl<'g> FeatureId<'g> {
         self.package_id
     }
 
-    /// Returns the name of the feature, or `None` if this is the "base" feature for this package.
-    pub fn feature(&self) -> Option<&'g str> {
-        self.feature
+    /// Returns the [`FeatureLabel`] associated with the feature.
+    pub fn label(&self) -> FeatureLabel<'g> {
+        self.label
     }
 
-    /// Returns true if this is the "base" feature for the package.
+    /// Returns true if this is the base feature for the package.
+    #[inline]
     pub fn is_base(&self) -> bool {
-        self.feature.is_none()
+        self.label.kind().is_base()
+    }
+
+    /// Returns true if this is an optional dependency.
+    #[inline]
+    pub fn is_optional_dependency(self) -> bool {
+        self.label.kind().is_optional_dependency()
+    }
+
+    /// Returns true if this is a named feature.
+    #[inline]
+    pub fn is_named(self) -> bool {
+        self.label.kind().is_named()
     }
 }
 
-impl<'g> From<(&'g PackageId, &'g str)> for FeatureId<'g> {
-    fn from((package_id, feature): (&'g PackageId, &'g str)) -> Self {
-        FeatureId::new(package_id, feature)
+impl<'g> From<(&'g PackageId, FeatureLabel<'g>)> for FeatureId<'g> {
+    fn from((package_id, label): (&'g PackageId, FeatureLabel<'g>)) -> Self {
+        FeatureId { package_id, label }
     }
 }
 
-impl<'g> From<(&'g PackageId, Option<&'g str>)> for FeatureId<'g> {
-    fn from((package_id, feature): (&'g PackageId, Option<&'g str>)) -> Self {
-        FeatureId {
-            package_id,
-            feature,
-        }
-    }
-}
-
-impl<'g> From<FeatureId<'g>> for (PackageId, Option<String>) {
-    fn from(feature_id: FeatureId<'g>) -> Self {
-        (
-            feature_id.package_id().clone(),
-            feature_id.feature().map(|feature| feature.to_string()),
-        )
-    }
-}
-
-/// The `Display` impl prints out `{package id}/feature`, or `{package id}/[base]`.
+/// The `Display` impl prints out:
+///
+/// * `{package-id}/[base]` for base features.
+/// * `{package-id}/feature-name` for named features.
+/// * `{package-id}/dep:dep-name` for optional dependencies.
 ///
 /// ## Examples
 ///
@@ -475,19 +512,81 @@ impl<'g> From<FeatureId<'g>> for (PackageId, Option<String>) {
 /// let package_id = PackageId::new("region 2.1.2 (registry+https://github.com/rust-lang/crates.io-index)");
 ///
 /// assert_eq!(
-///     format!("{}", FeatureId::new(&package_id, "foo")),
+///     format!("{}", FeatureId::base(&package_id)),
+///     "region 2.1.2 (registry+https://github.com/rust-lang/crates.io-index)/[base]"
+/// );
+///
+/// assert_eq!(
+///     format!("{}", FeatureId::named(&package_id, "foo")),
 ///     "region 2.1.2 (registry+https://github.com/rust-lang/crates.io-index)/foo"
 /// );
 ///
 /// assert_eq!(
-///     format!("{}", FeatureId::base(&package_id)),
-///     "region 2.1.2 (registry+https://github.com/rust-lang/crates.io-index)/[base]"
+///     format!("{}", FeatureId::optional_dependency(&package_id, "bar")),
+///     "region 2.1.2 (registry+https://github.com/rust-lang/crates.io-index)/dep:bar"
 /// );
 /// ```
 impl<'g> fmt::Display for FeatureId<'g> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        let feature_name = self.feature.unwrap_or("[base]");
-        write!(f, "{}/{}", self.package_id, feature_name)
+        write!(f, "{}/{}", self.package_id, self.label)
+    }
+}
+
+/// A unique identifier for a feature within a specific package. Forms part of a [`FeatureId`].
+#[derive(Clone, Copy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub enum FeatureLabel<'g> {
+    /// The "base" feature. Every package has one such feature.
+    Base,
+
+    /// This is a named feature in the `[features]` section, or an implicit feature that corresponds to
+    /// an optional dependency.
+    ///
+    /// For versions of Cargo prior to 1.60, optional dependencies always create implicit features
+    /// by the same name. For versions 1.60 and greater, optional dependencies may create implicit
+    /// features if the dependency doesn't exist with the name "dep" in it.
+    Named(&'g str),
+
+    /// This is an optional dependency.
+    OptionalDependency(&'g str),
+}
+
+impl<'g> FeatureLabel<'g> {
+    /// Returns the kind of feature this is.
+    ///
+    /// The kind of a feature is simply the enum variant without any associated data.
+    #[inline]
+    pub fn kind(self) -> FeatureKind {
+        match self {
+            Self::Base => FeatureKind::Base,
+            Self::Named(_) => FeatureKind::Named,
+            Self::OptionalDependency(_) => FeatureKind::OptionalDependency,
+        }
+    }
+}
+
+/// The `Display` impl for `FeatureLabel` prints out:
+///
+/// * `[base]` for base labels.
+/// * `feature-name` for optional dependencies.
+/// * `dep:dep-name` for named features.
+///
+/// ## Examples
+///
+/// ```
+/// use guppy::graph::feature::FeatureLabel;
+///
+/// assert_eq!(format!("{}", FeatureLabel::Base), "[base]");
+/// assert_eq!(format!("{}", FeatureLabel::Named("foo")), "foo");
+/// assert_eq!(format!("{}", FeatureLabel::OptionalDependency("bar")), "dep:bar");
+/// ```
+
+impl<'g> fmt::Display for FeatureLabel<'g> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::Base => write!(f, "[base]"),
+            Self::Named(feature_name) => write!(f, "{}", feature_name),
+            Self::OptionalDependency(dep_name) => write!(f, "dep:{}", dep_name),
+        }
     }
 }
 
@@ -520,9 +619,9 @@ impl<'g> FeatureMetadata<'g> {
             .expect("valid package ID")
     }
 
-    /// Returns the type of this feature.
-    pub fn feature_type(&self) -> FeatureType {
-        self.inner.feature_type
+    /// Returns the label for this feature.
+    pub fn label(&self) -> FeatureLabel<'g> {
+        self.feature_id().label()
     }
 
     // ---
@@ -543,13 +642,14 @@ impl<'g> FeatureMetadata<'g> {
 /// A graph representing every possible feature of every package, and the connections between them.
 #[derive(Clone, Debug)]
 pub(in crate::graph) struct FeatureGraphImpl {
-    pub(super) graph: Graph<FeatureNode, FeatureEdge, Directed, FeatureIx>,
+    pub(super) graph: FeaturePetgraph,
     // base ixs consists of the base (start) feature indexes for each package.
     pub(super) base_ixs: Vec<NodeIndex<FeatureIx>>,
     pub(super) map: HashMap<FeatureNode, FeatureMetadataImpl>,
     pub(super) warnings: Vec<FeatureGraphWarning>,
     // The strongly connected components of the feature graph. Computed on demand.
     pub(super) sccs: OnceCell<Sccs<FeatureIx>>,
+    pub(super) weak: WeakDependencies,
 }
 
 impl FeatureGraphImpl {
@@ -720,7 +820,7 @@ impl<'g> ConditionalLink<'g> {
     // ---
 
     #[allow(dead_code)]
-    pub(self) fn edge_ix(&self) -> EdgeIndex<FeatureIx> {
+    pub(super) fn edge_ix(&self) -> EdgeIndex<FeatureIx> {
         self.edge_ix
     }
 
@@ -736,30 +836,14 @@ impl<'g> ConditionalLink<'g> {
 #[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
 pub(in crate::graph) struct FeatureNode {
     package_ix: NodeIndex<PackageIx>,
-    feature_idx: Option<usize>,
+    feature_idx: FeatureIndexInPackage,
 }
 
 impl FeatureNode {
     /// Returns a new feature node.
-    pub(in crate::graph) fn new(package_ix: NodeIndex<PackageIx>, feature_idx: usize) -> Self {
-        Self {
-            package_ix,
-            feature_idx: Some(feature_idx),
-        }
-    }
-
-    /// Returns a new feature node representing the base package with no features enabled.
-    pub(in crate::graph) fn base(package_ix: NodeIndex<PackageIx>) -> Self {
-        Self {
-            package_ix,
-            feature_idx: None,
-        }
-    }
-
-    /// Returns a new feature node, can also be the base.
-    pub(in crate::graph) fn new_opt(
+    pub(in crate::graph) fn new(
         package_ix: NodeIndex<PackageIx>,
-        feature_idx: Option<usize>,
+        feature_idx: FeatureIndexInPackage,
     ) -> Self {
         Self {
             package_ix,
@@ -767,23 +851,57 @@ impl FeatureNode {
         }
     }
 
+    /// Base feature node.
+    pub(in crate::graph) fn base(package_ix: NodeIndex<PackageIx>) -> Self {
+        Self {
+            package_ix,
+            feature_idx: FeatureIndexInPackage::Base,
+        }
+    }
+
+    pub(in crate::graph) fn optional_dep(package_ix: NodeIndex<PackageIx>, dep_idx: usize) -> Self {
+        Self {
+            package_ix,
+            feature_idx: FeatureIndexInPackage::OptionalDependency(dep_idx),
+        }
+    }
+
+    pub(in crate::graph) fn named_feature(
+        package_ix: NodeIndex<PackageIx>,
+        named_idx: usize,
+    ) -> Self {
+        Self {
+            package_ix,
+            feature_idx: FeatureIndexInPackage::Named(named_idx),
+        }
+    }
+
     fn from_id(feature_graph: &FeatureGraph<'_>, id: FeatureId<'_>) -> Option<Self> {
         let metadata = feature_graph.package_graph.metadata(id.package_id()).ok()?;
-        match id.feature() {
-            Some(feature_name) => Some(FeatureNode::new(
-                metadata.package_ix(),
-                metadata.get_feature_idx(feature_name)?,
-            )),
-            None => Some(FeatureNode::base(metadata.package_ix())),
-        }
+        Some(FeatureNode::new(
+            metadata.package_ix(),
+            metadata.get_feature_idx(id.label())?,
+        ))
     }
 
     pub(super) fn named_features(package: PackageMetadata<'_>) -> impl Iterator<Item = Self> + '_ {
         let package_ix = package.package_ix();
-        package.named_features_full().map(move |(n, _, _)| Self {
-            package_ix,
-            feature_idx: Some(n),
-        })
+        package
+            .named_features_full()
+            .map(move |(feature_idx, _, _)| Self {
+                package_ix,
+                feature_idx,
+            })
+    }
+
+    pub(super) fn optional_deps(package: PackageMetadata<'_>) -> impl Iterator<Item = Self> + '_ {
+        let package_ix = package.package_ix();
+        package
+            .optional_deps_full()
+            .map(move |(feature_idx, _)| Self {
+                package_ix,
+                feature_idx,
+            })
     }
 
     pub(in crate::graph) fn package_ix(&self) -> NodeIndex<PackageIx> {
@@ -799,27 +917,54 @@ impl FeatureNode {
 pub enum FeatureEdge {
     /// This edge is from a feature to its base package.
     FeatureToBase,
-    /// This is a conditional edge, e.g. through:
+
+    /// This is a dependency edge, e.g.:
     ///
     /// ```toml
     /// [dependencies]
     /// foo = { version = "1", features = ["a", "b"] }
     /// ```
     ///
-    /// or through
+    /// (The above is conditional in that it isn't a build dependency. Similarly, it could be
+    /// a target-specific dependency.)
+    ///
+    /// This also includes optional dependencies, for which the "from" node is
+    /// `FeatureLabel::OptionalDependency` rather than `FeatureLabel::Base`.
     ///
     /// ```toml
-    /// [features]
-    /// "a" = ["foo/b"]
+    /// [dependencies]
+    /// foo = { version = "1", features = ["a", "b"], optional = true }
     /// ```
-    Conditional(ConditionalLinkImpl),
+    DependenciesSection(ConditionalLinkImpl),
+
     /// This edge is from a feature depending on other features within the same package:
     ///
     /// ```toml
     /// [features]
-    /// "a" = ["b"]
+    /// a = ["b"]
     /// ```
-    FeatureDependency,
+    NamedFeature,
+
+    /// This edge is from a feature to an optional dependency.
+    ///
+    /// ```toml
+    /// [features]
+    /// a = ["dep:foo"]
+    /// ```
+    NamedFeatureDepColon(ConditionalLinkImpl),
+
+    /// This is a named feature line of the form
+    ///
+    /// ```toml
+    /// [features]
+    /// a = ["foo/b"]
+    /// # or
+    /// a = ["foo?/b"]
+    /// ```
+    NamedFeatureWithSlash {
+        link: ConditionalLinkImpl,
+        weak_index: Option<WeakIndex>,
+    },
 }
 
 /// Not part of the stable API -- only exposed for FeatureSet::links().
@@ -843,18 +988,44 @@ impl ConditionalLinkImpl {
 #[derive(Clone, Debug, Eq, Hash, PartialEq)]
 pub(super) struct FeatureMetadataImpl {
     pub(super) feature_ix: NodeIndex<FeatureIx>,
-    pub(super) feature_type: FeatureType,
 }
 
-/// The type of a particular feature within a package.
+/// The kind of a particular feature within a package.
 ///
 /// Returned by `FeatureMetadata`.
 #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
-pub enum FeatureType {
-    /// This is a named feature in the `[features]` section.
-    NamedFeature,
+pub enum FeatureKind {
+    /// The "base" feature. Every package has one such feature.
+    Base,
+
+    /// This is a named feature in the `[features]` section, or an implicit feature that corresponds to
+    /// an optional dependency.
+    ///
+    /// For versions of Cargo prior to 1.60, optional dependencies always create implicit features
+    /// by the same name. For versions 1.60 and greater, optional dependencies may create implicit
+    /// features if the dependency doesn't exist with the name "dep" in it.
+    Named,
+
     /// This is an optional dependency.
-    OptionalDep,
-    /// This is the "base" package with no features enabled.
-    BasePackage,
+    OptionalDependency,
+}
+
+impl FeatureKind {
+    /// Returns true if this is the base feature.
+    #[inline]
+    pub fn is_base(self) -> bool {
+        matches!(self, Self::Base)
+    }
+
+    /// Returns true if this is a named feature.
+    #[inline]
+    pub fn is_named(self) -> bool {
+        matches!(self, Self::Named)
+    }
+
+    /// Returns true if this is an optional dependency.
+    #[inline]
+    pub fn is_optional_dependency(self) -> bool {
+        matches!(self, Self::OptionalDependency)
+    }
 }
