@@ -15,9 +15,12 @@ mod graph_impl;
 mod proptest_helpers;
 mod query;
 mod resolve;
+mod weak;
 
+pub(self) use build::*;
 pub use cycles::*;
 pub use feature_list::FeatureList;
 pub use graph_impl::*;
 pub use query::*;
 pub use resolve::*;
+pub use weak::*;
