@@ -6,17 +6,18 @@ use crate::{
     graph::{
         cargo::{CargoOptions, CargoSet},
         feature::{
-            ConditionalLink, FeatureEdge, FeatureGraph, FeatureId, FeatureList, FeatureMetadata,
-            FeatureQuery, FeatureResolver,
+            build::FeatureEdgeReference, ConditionalLink, FeatureEdge, FeatureGraph, FeatureId,
+            FeatureList, FeatureMetadata, FeatureQuery, FeatureResolver,
         },
         resolve_core::ResolveCore,
         DependencyDirection, FeatureGraphSpec, FeatureIx, PackageIx, PackageMetadata, PackageSet,
     },
-    petgraph_support::IxBitSet,
+    petgraph_support::{dfs::BufferedEdgeFilterFn, IxBitSet},
     sorted_set::SortedSet,
     Error, PackageId,
 };
 use fixedbitset::FixedBitSet;
+use itertools::Either;
 use petgraph::{graph::NodeIndex, visit::EdgeRef};
 
 impl<'g> FeatureGraph<'g> {
@@ -84,23 +85,36 @@ impl<'g> FeatureSet<'g> {
     ) -> Self {
         let graph = query.graph;
         let params = query.params.clone();
-        Self {
-            graph,
-            core: ResolveCore::with_edge_filter(graph.dep_graph(), params, |edge| {
-                match graph.edge_to_conditional_link(
-                    edge.source(),
-                    edge.target(),
-                    edge.id(),
-                    Some(edge.weight()),
-                ) {
-                    Some(conditional_link) => resolver.accept(&query, conditional_link),
-                    None => {
-                        // Feature links within the same package are always followed.
-                        true
-                    }
+
+        // State used by the callback below.
+        let mut buffer_states = graph
+            .inner
+            .weak
+            .new_buffer_states(|link| resolver.accept(&query, link));
+
+        let filter_fn = |edge_ref: FeatureEdgeReference<'g>| {
+            match graph.edge_to_conditional_link(
+                edge_ref.source(),
+                edge_ref.target(),
+                edge_ref.id(),
+                Some(edge_ref.weight()),
+            ) {
+                Some((link, weak_index)) => buffer_states.track(edge_ref, link, weak_index),
+                None => {
+                    // Feature links within the same package are always followed.
+                    Either::Left(Some(edge_ref))
                 }
-            }),
-        }
+            }
+            .into_iter()
+        };
+
+        let core = ResolveCore::with_buffered_edge_filter(
+            graph.dep_graph(),
+            params,
+            BufferedEdgeFilterFn(filter_fn),
+        );
+
+        Self { graph, core }
     }
 
     #[allow(dead_code)]
@@ -504,7 +518,9 @@ impl<'g> FeatureSet<'g> {
         self.core
             .links(graph.dep_graph(), graph.sccs(), direction)
             .filter_map(move |(source_ix, target_ix, edge_ix)| {
-                graph.edge_to_conditional_link(source_ix, target_ix, edge_ix, None)
+                graph
+                    .edge_to_conditional_link(source_ix, target_ix, edge_ix, None)
+                    .map(|(link, _)| link)
             })
     }
 
