@@ -67,9 +67,12 @@ impl<'g> FeatureSet<'g> {
                 let info = PackageInfo {
                     status,
                     features: feature_list
-                        .features()
-                        .iter()
-                        .map(|feature| feature.to_string())
+                        .named_features()
+                        .map(|feature| feature.to_owned())
+                        .collect(),
+                    optional_deps: feature_list
+                        .optional_deps()
+                        .map(|dep| dep.to_owned())
                         .collect(),
                 };
 
@@ -174,9 +177,12 @@ impl CargoOptionsSummary {
             .map(|features| FeaturesOnlySummary {
                 summary_id: features.package().to_summary_id(),
                 features: features
-                    .features()
-                    .iter()
-                    .map(|&feature| feature.to_owned())
+                    .named_features()
+                    .map(|feature| feature.to_owned())
+                    .collect(),
+                optional_deps: features
+                    .optional_deps()
+                    .map(|feature| feature.to_owned())
                     .collect(),
             })
             .collect::<Vec<_>>();
@@ -273,8 +279,12 @@ pub struct FeaturesOnlySummary {
     #[serde(flatten)]
     pub summary_id: SummaryId,
 
-    /// The features built for this package.
+    /// The named features built for this package.
     pub features: BTreeSet<String>,
+
+    /// The optional dependencies built for this package.
+    #[serde(skip_serializing_if = "BTreeSet::is_empty", default)]
+    pub optional_deps: BTreeSet<String>,
 }
 
 impl<'g> PackageSource<'g> {
